import aio_pika.message
import pydantic
import time
import uuid

from aio_pika.message import IncomingMessage
from dataclasses import dataclass
from enum import IntEnum
from typing import Dict, Any, Optional, TypeVar, Union

from .models import BaseModel


__all__ = [
    "AbstractOptions",
    "IncomingMessage",
    "Message",
    "Priority",
    "RequestContext",
]


TAbstractOptions = TypeVar("TAbstractOptions", bound="AbstractOptions")


class AbstractOptions(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid", frozen=True)

    def __init__(self, *args: "AbstractOptions", **kwargs: Dict[str, Any]) -> None:
        if args:
            merged = {}
            self_class = type(self)
            for options in args:
                if not isinstance(options, self_class):
                    raise TypeError(
                        f"Positional arguments must be {self_class} instances"
                    )

                merged.update(options.model_dump(exclude_unset=True))
            merged.update(**kwargs)

            kwargs = merged

        return super().__init__(**kwargs)

    def update(
        self: TAbstractOptions, *args: "TAbstractOptions", **kwargs: Any
    ) -> "TAbstractOptions":
        if not args and not kwargs:
            return self

        return self.__class__(self, *args, **kwargs)


class Priority(IntEnum):
    LOW = 0
    NORMAL = 1
    DEFAULT = 1
    HIGH = 2
    INTERACTIVE = 3
    SYSTEM = 4


class Message(aio_pika.message.Message):
    """AMQP message abstraction"""

    __slots__ = ()

    def __init__(
        self,
        body: Union[bytes, BaseModel],
        *,
        headers: Optional[aio_pika.message.HeadersType] = None,
        content_type: Optional[str] = None,
        content_encoding: Optional[str] = None,
        delivery_mode: Union[aio_pika.message.DeliveryMode, int, None] = None,
        priority: Optional[int] = None,
        correlation_id: Optional[str] = None,
        reply_to: Optional[str] = None,
        expiration: Optional[aio_pika.message.DateType] = None,
        message_id: Optional[str] = None,
        timestamp: Optional[aio_pika.message.DateType] = None,
        type: Optional[str] = None,
        user_id: Optional[str] = None,
        app_id: Optional[str] = None,
    ) -> None:
        if isinstance(body, BaseModel):
            body = body.model_dump_msgpack()
            content_type = None

        super().__init__(
            body=body,
            headers=headers,
            content_type=content_type or "application/msgpack",
            content_encoding=content_encoding,
            delivery_mode=delivery_mode or aio_pika.DeliveryMode.PERSISTENT,
            priority=priority,
            correlation_id=correlation_id,
            reply_to=reply_to,
            expiration=expiration,
            message_id=message_id or str(uuid.uuid1()),
            timestamp=timestamp or time.time(),
            type=type,
            user_id=user_id,
            app_id=app_id,
        )


@dataclass(frozen=True)
class RequestContext:
    deadline: Optional[float]
    priority: Priority



import datetime
import functools
import inspect
import logging
import re

from aio_pika.message import Message

from typing import Any, Callable


logger = logging.getLogger(__name__)


def fix_message_timestamp(func: Callable) -> Callable:
    @functools.wraps(func)
    def inner(message: Message) -> Any:
        # Fix `pamqp` naive timestamp
        if message.timestamp:
            message.timestamp = message.timestamp.replace(tzinfo=datetime.timezone.utc)

        return func(message)

    return inner


def clean_service_name(service_name: str) -> str:
    return re.sub("[^a-z0-9-]+", "-", service_name.strip().lower())


def assert_handler_is_coroutine(handler: Callable) -> None:
    if not inspect.iscoroutinefunction(handler):
        raise TypeError(f"{handler!r} must be coroutine functions")


def assert_handler_kwargs_only(handler: Callable) -> None:
    positional_only = []
    positional_or_keyword = []

    for parameter_name, parameter in inspect.signature(handler).parameters.items():
        if parameter.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            positional_or_keyword.append(parameter_name)

        elif parameter.kind in [
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.VAR_POSITIONAL,
        ]:
            positional_only.append(parameter_name)

    if positional_only:
        raise TypeError(
            f"{handler!r} has positional-only parameters {positional_only} that are not supported"
        )

    if positional_or_keyword:
        logger.warning(
            "%s has positional parameters %s, only keyword parameters are supported",
            handler,
            positional_or_keyword,
        )



class RpcError(Exception):
    pass


class RetryableError(RpcError):
    pass


class UnknownError(RpcError):
    pass


class UnhandledError(RpcError):
    pass


class ParsingError(RpcError):
    pass


class UnknownServiceError(RpcError):
    pass


class UnknownMethodError(RpcError):
    pass


class ExecutionError(RpcError):
    pass


class RetryableExecutionError(RetryableError):
    pass



import aio_pika.message
import functools
import logging
import msgpack
import pydantic
import zlib

from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar

from . import errors


logger = logging.getLogger(__name__)


TBaseModel = TypeVar("TBaseModel", bound="BaseModel")


class BaseModel(pydantic.BaseModel):
    @classmethod
    def model_validate_message(
        cls: Type[TBaseModel], message: aio_pika.message.Message
    ) -> "TBaseModel":
        body = message.body
        if message.content_encoding == "deflate":
            try:
                body = zlib.decompress(body)
            except Exception as error:
                raise errors.ParsingError(f"Message decompression failed: `{error!r}`")

        elif message.content_encoding:
            raise errors.ParsingError(
                f"Unknown content_encoding: `{message.content_encoding}`"
            )

        if message.content_type != "application/msgpack":
            raise errors.ParsingError(
                f"Got a message with unknown content type: {message.content_type}"
            )

        try:
            data = msgpack.loads(body, timestamp=3, ext_hook=msgpack_ext_hook)
            return cls.model_validate(data)
        except ValueError as error:
            raise errors.ParsingError(f"Message deserialization failed: `{error!r}`")

    def model_dump_msgpack(self, **kwargs: Any) -> bytes:
        kwargs.setdefault("exclude_defaults", True)
        data = self.model_dump(**kwargs)
        return msgpack.dumps(data, datetime=True, default=msgpack_default)


class Request(BaseModel):
    method_name: str
    arguments: Dict[str, Any] = {}
    context: Optional[Any] = None
    timings: List[Tuple[str, float]] = []


class ResponseError(BaseModel):
    type: str
    message: Optional[str] = None


class Response(BaseModel):
    result: Optional[Any] = None
    error: Optional[ResponseError] = None
    context: Optional[Any] = None
    timings: List[Tuple[str, float]] = []

    @classmethod
    def wrap_errors(cls, func: Callable) -> Callable:
        @functools.wraps(func)
        async def inner(*args: Any, **kwargs: Any) -> Any:
            try:
                response = await func(*args, **kwargs)
                if not isinstance(response, cls):
                    raise TypeError(f"Incorrect response type, got: {type(response)}")

            except errors.RpcError as error:
                response = cls(
                    error=ResponseError(
                        type=type(error).__name__, message=error.args[0]
                    )
                )

            except Exception as error:
                logger.exception("Unhandled error")
                response = cls(
                    error=ResponseError(type="UnhandledError", message=repr(error))
                )

            return response

        return inner


ObjectId: Optional[Type] = None
try:
    from bson import ObjectId
except ImportError:
    pass


def msgpack_default(obj: Any) -> msgpack.ExtType:
    if ObjectId is not None and isinstance(obj, ObjectId):
        return msgpack.ExtType(1, obj.binary)

    raise TypeError(f"Cannot serialize {obj!r}")


def msgpack_ext_hook(code: int, data: bytes) -> Any:
    if code == 1:
        if ObjectId is None:
            raise RuntimeError("Install `bson` package to support `ObjectId` type")

        return ObjectId(data)

    return msgpack.ExtType(code, data)



from __future__ import annotations

import aio_pika.exceptions
import asyncio
import importlib.metadata
import logging
import weakref

from aio_pika.connection import URL
from aio_pika.robust_connection import AbstractRobustConnection
from typing import Union


logger = logging.getLogger(__name__)

_connect_lock = asyncio.Lock()
_connections: weakref.WeakValueDictionary[URL, AbstractRobustConnection] = (
    weakref.WeakValueDictionary()
)


async def get_connection(
    connection_string: Union[str, URL, None], service_name: str = "?"
) -> AbstractRobustConnection:
    connection_url = aio_pika.connection.make_url(connection_string)
    if not connection_url.query.get("name"):
        package_name = __package__.split(".")[0]
        package_name_version = importlib.metadata.version(package_name)

        connection_url %= {
            "name": f"{service_name} ({package_name} v{package_name_version})"
        }

    async with _connect_lock:
        connection = _connections.get(connection_url)
        if connection is None:
            connection = await aio_pika.connect_robust(connection_url)
            _connections[connection_url] = connection

    return connection
